@@ -1,11 +1,14 @@
 #!/usr/bin/env python3
 from snowfakery.generate_mapping_from_recipe import mapping_from_recipe_templates
+from snowfakery.salesforce import create_cci_record_type_tables
 from snowfakery.output_streams import (
     DebugOutputStream,
-    SqlOutputStream,
+    SqlDbOutputStream,
+    SqlTextOutputStream,
     JSONOutputStream,
     CSVOutputStream,
     ImageOutputStream,
+    GraphvizOutputStream,
     MultiplexOutputStream,
 )
 from snowfakery.data_gen_exceptions import DataGenError
@@ -23,9 +26,7 @@ if __name__ == "__main__":  # pragma: no cover
     sys.path.append(str(Path(__file__).parent.parent))
 
 
-file_extensions = [
-    "JSON",
-    "json",
+graphic_file_extensions = [
     "PNG",
     "png",
     "SVG",
@@ -35,9 +36,15 @@ file_extensions = [
     "jpg",
     "ps",
     "dot",
+]
+
+file_extensions = [
+    "JSON",
+    "json",
     "txt",
     "csv",
-]
+    "sql",
+] + graphic_file_extensions
 
 
 def eval_arg(arg):
@@ -71,6 +78,8 @@ def int_string_tuple(ctx, param, value=None):
 
 
 @click.command()
+# TODO: This should become type=click.File("r")
+#       For consistency and flexibility
 @click.argument("yaml_file", type=click.Path(exists=True))
 @click.option(
     "--dburl",
@@ -102,8 +111,8 @@ def int_string_tuple(ctx, param, value=None):
 @click.option("--cci-mapping-file", "mapping_file", type=click.Path(exists=True))
 @click.option(
     "--generate-cci-mapping-file",
-    "generate_cci_mapping_file",
-    type=click.Path(exists=False),
+    type=click.File("w"),
+    help="Generate a CumulusCI mapping file for the dataset",
 )
 @click.option(
     "--generate-continuation-file",
@@ -111,6 +120,12 @@ def int_string_tuple(ctx, param, value=None):
     help="A file that captures information about how to continue a "
     "multi-batch data generation process",
 )
+@click.option(
+    "--generate-record-type-tables/--no-generate-record-type-tables",
+    "should_create_cci_record_type_tables",
+    default=False,
+    hidden=True,
+)
 @click.option(
     "--continuation-file",
     type=click.File("r"),
@@ -128,9 +143,9 @@ def generate_cli(
     output_format=None,
     output_files=None,
     output_folder=None,
-    nickname_ids=None,
     continuation_file=None,
     generate_continuation_file=None,
+    should_create_cci_record_type_tables=False,
 ):
     """
         Generates records from a YAML file
@@ -138,12 +153,12 @@ def generate_cli(
     \b
         Records can go to:
             * stdout (default)
-            * JSON file (--output_format=json --output-file=foo.json)
-            * diagram file (--output_format=png --output-file=foo.png)
+            * JSON file (--output-format=json --output-file=foo.json)
+            * diagram file (--output-format=png --output-file=foo.png)
             * a database identified by --dburl (e.g. --dburl sqlite:////tmp/foo.db)
             * or to a directory as a set of CSV files (--output-format=csv --output-folder=csvfiles)
 
-        Diagram output depends on the installation of pygraphviz ("pip install pygraphviz")
+        Diagram output depends on the installation of graphviz (https://www.graphviz.org/download/)
 
         Full documentation here:
 
@@ -182,10 +197,11 @@ def generate_cli(
                 )
                 sys.stderr.write(debuginfo)
             if generate_cci_mapping_file:
-                with click.open_file(generate_cci_mapping_file, "w") as f:
-                    yaml.safe_dump(
-                        mapping_from_recipe_templates(summary), f, sort_keys=False
-                    )
+                yaml.safe_dump(
+                    mapping_from_recipe_templates(summary),
+                    generate_cci_mapping_file,
+                    sort_keys=False,
+                )
         except DataGenError as e:
             if debug_internals:
                 raise e
@@ -193,6 +209,8 @@ def generate_cli(
                 click.echo("")
                 click.echo(e.prefix)
                 raise click.ClickException(str(e)) from e
+    if should_create_cci_record_type_tables:
+        create_cci_record_type_tables(dburls[0])
 
 
 @contextmanager
@@ -209,12 +227,15 @@ def configure_output_stream(
         else:
             mappings = None
 
-        output_streams.append(SqlOutputStream.from_url(dburl, mappings))
+        output_streams.append(SqlDbOutputStream.from_url(dburl, mappings))
 
-    # JSON is the only output format (other than debug) that can go on stdout
+    # JSON and SQL are the only output formats (other than debug) that can go on stdout
     if output_format == "json" and not output_files:
         output_streams.append(JSONOutputStream(sys.stdout))
 
+    if output_format == "sql" and not output_files:
+        output_streams.append(SqlTextOutputStream(sys.stdout))
+
     if output_format == "csv":
         output_streams.append(CSVOutputStream(output_folder))
 
@@ -226,9 +247,13 @@ def configure_output_stream(
 
             if format == "json":
                 output_streams.append(JSONOutputStream(path))
+            elif format == "sql":
+                output_streams.append(SqlTextOutputStream(path))
             elif format == "txt":
                 output_streams.append(DebugOutputStream(path))
-            elif format in file_extensions:
+            elif format == "dot":
+                output_streams.append(GraphvizOutputStream(path))
+            elif format in graphic_file_extensions:
                 output_streams.append(ImageOutputStream(path, format))
             else:
                 raise click.ClickException(
@@ -245,6 +270,7 @@ def configure_output_stream(
         yield output_stream
     finally:
         for output_stream in output_streams:
+            messages = None
             try:
                 messages = output_stream.close()
             except Exception as e:
