@@ -1,10 +1,6 @@
 from abc import abstractmethod, ABC
-from .data_generator_runtime import (
-    evaluate_function,
-    ObjectRow,
-    RuntimeContext,
-    NicknameSlot,
-)
+from .data_generator_runtime import evaluate_function, RuntimeContext, Interpreter
+from .object_rows import ObjectRow, ObjectReference
 from contextlib import contextmanager
 from typing import Union, Dict, Sequence, Optional, cast
 from .utils.template_utils import look_for_number
@@ -23,7 +19,7 @@ from .plugins import Scalar, PluginResult
 # objects that represent the hierarchy of a data generator.
 # roughly similar to the YAML structure but with domain-specific objects
 Definition = Union["ObjectTemplate", "SimpleValue", "StructuredValue"]
-FieldValue = Union[None, Scalar, ObjectRow, tuple, NicknameSlot, PluginResult]
+FieldValue = Union[None, Scalar, ObjectRow, tuple, PluginResult, ObjectReference]
 
 
 class FieldDefinition(ABC):
@@ -45,12 +41,38 @@ class FieldDefinition(ABC):
         try:
             yield
         except DataGenError as e:
-            raise fix_exception(message, self, e)
+            raise fix_exception(message, self, e) from e
         except Exception as e:
             message = message.format(*args, **kwargs)
             raise DataGenError(f"{message} : {str(e)}", self.filename, self.line_num)
 
 
+class VariableDefinition:
+    """Defines a mutable variable"""
+
+    tablename = None
+
+    def __init__(
+        self, filename: str, line_num: int, varname: str, expression: Definition
+    ):
+        self.varname = varname
+        self.expression = expression
+        self.filename = filename
+        self.line_num = line_num
+
+    def evaluate(self, context: RuntimeContext) -> FieldValue:
+        """Evaluate the expression"""
+        return self.expression.render(context)
+
+    def execute(
+        self, interp: Interpreter, parent_context: RuntimeContext, continuing: bool
+    ) -> Optional[dict]:
+        with parent_context.child_context(self) as context:
+            name = self.varname
+            value = self.evaluate(context)
+        interp.register_variable(name, value)
+
+
 class ObjectTemplate:
     """A factory that generates rows.
 
@@ -87,7 +109,7 @@ class ObjectTemplate:
         return self.generate_rows(context.output_stream, context)
 
     def generate_rows(
-        self, storage, parent_context: RuntimeContext
+        self, output_stream, parent_context: RuntimeContext
     ) -> Optional[ObjectRow]:
         """Generate several rows"""
         rc = None
@@ -95,7 +117,7 @@ class ObjectTemplate:
             count = self._evaluate_count(context)
             with self.exception_handling(f"Cannot generate {self.name}"):
                 for i in range(count):
-                    rc = self._generate_row(storage, context, i)
+                    rc = self._generate_row(output_stream, context, i)
 
         return rc  # return last row
 
@@ -126,30 +148,31 @@ class ObjectTemplate:
     def name(self) -> str:
         name = self.tablename
         if self.nickname:
-            name += " (self.nickname)"
+            name += f" ({self.nickname})"
         return name
 
     @property
     def id(self):
         return id(self)
 
-    def _generate_row(self, storage, context: RuntimeContext, index: int) -> ObjectRow:
+    def _generate_row(
+        self, output_stream, context: RuntimeContext, index: int
+    ) -> ObjectRow:
         """Generate an individual row"""
         id = context.generate_id(self.nickname)
         row = {"id": id}
         sobj = ObjectRow(self.tablename, row, index)
 
-        context.register_object(sobj, self.nickname)
+        context.register_object(sobj, self.nickname, self.just_once)
 
         self._generate_fields(context, row)
 
         with self.exception_handling("Cannot write row"):
             self.register_row_intertable_references(row, context)
             if not self.tablename.startswith("__"):
-                storage.write_row(self.tablename, row)
+                output_stream.write_row(self.tablename, row)
 
-        for i, childobj in enumerate(self.friends):
-            childobj.generate_rows(storage, context)
+        context.interpreter.loop_over_templates_once(self.friends, True)
         return sobj
 
     def _generate_fields(self, context: RuntimeContext, row: Dict) -> None:
@@ -173,11 +196,18 @@ class ObjectTemplate:
     ) -> None:
         """Before serializing we need to convert objects to flat ID integers."""
         for fieldname, fieldvalue in row.items():
-            if isinstance(fieldvalue, ObjectRow):
+            if isinstance(fieldvalue, (ObjectRow, ObjectReference)):
                 context.register_intertable_reference(
                     self.tablename, fieldvalue._tablename, fieldname
                 )
 
+    def execute(
+        self, interp: Interpreter, context: RuntimeContext, continuing: bool
+    ) -> Optional[ObjectRow]:
+        should_skip = self.just_once and continuing
+        if not should_skip:
+            self.generate_rows(interp.output_stream, interp.current_context)
+
 
 class SimpleValue(FieldDefinition):
     """A value with no sub-structure (although it could hold a template)
@@ -335,7 +365,10 @@ class FieldFactory:
         except Exception as e:
             raise fix_exception(
                 f"Problem rendering field {self.name}:\n {str(e)}", self, e
-            )
+            ) from e
 
     def __repr__(self):
         return f"<{self.__class__.__name__, self.name, self.definition.__class__.__name__}>"
+
+
+Statement = Union[ObjectTemplate, VariableDefinition]
